/*
 *
 * Pocket Worm
 *     by Sebastian Mihai, 2015
 *     http://sebastianmihai.com
 *
 */

#define	CLOCK_SPEED	CLOCK_8M

#include <pocketlib.h>

#define byte short int
#define ubyte unsigned short int
#define ulong unsigned long
#define SCREEN_WIDTH 32
#define SCREEN_HEIGHT 25
#define PAD_ANY PAD_UP | PAD_DOWN | PAD_LEFT | PAD_RIGHT
#define MAX_WORM_LENGTH 128
#define NUM_FRAMES_DELAY 10
#define GOAL 30

typedef enum Outcome {
	Nothing,
	Loss,
	AteFood
} Outcome;

ulong g_frame = 0;

unsigned int g_nextRandom = 1337;
unsigned int random() {
    g_nextRandom = ((g_nextRandom * 14577) % 190) + 25;
    return g_nextRandom;
}

void delay() {
	g_frame++;
	VSync(0);
}

typedef struct Point {
	byte x, y;
} Point;

void menu() {
	// debounce from a victory
	while( ( PAD_STATUS & PAD_BUTTON ) ) { 
	}
	
	FastLCDClear();
	DrawString( 0, 0, " PUSH" );
	DrawString( 0, 6, "   FIRE" );
	DrawString( 0, 12, "TO START" );
	
	DrawString( 0, 20, " BY S.M." );
	DrawString( 0, 26, "    2015" );

	VSync(1);
	
	while( !( PAD_STATUS & PAD_BUTTON ) ) {
		g_frame++;
	}
}

void victory() {
	
	FastLCDClear();
	DrawString( 0, 9, "GOOD JOB" );
	DrawString( 0, 20, " PUSH" );
	DrawString( 0, 26, "   FIRE" );
	
	VSync(1);
	
	while( !( PAD_STATUS & PAD_BUTTON ) ) {
		g_frame++;	
	}
}

void display_worm( Point* worm, ubyte headIndex ) {
	ubyte c = 0;
	for( c = 0; c <= headIndex; c++ ) {
		DrawLCDPix( worm[c].x, worm[c].y );
	}
}

void display_food( Point* food ) {
	DrawLCDPix( food->x, food->y );
}

void move_worm( Point* worm, ubyte headIndex, byte dx, byte dy ) {
	ubyte c;
	// each worm segment moves to where the segment in front of it was
	for( c = 0; c < headIndex; c++ ) {
		worm[c].x = worm[c+1].x;
		worm[c].y = worm[c+1].y;
	}
	// and the head moves according to delta x and delta y
	worm[headIndex].x += dx;
	worm[headIndex].y += dy;
}

Outcome check_outcome( Point* worm, ubyte headIndex, Point food ) {
	ubyte c;
	
	if( worm[headIndex].x >= SCREEN_WIDTH ||
	    worm[headIndex].x < 0 ||
		worm[headIndex].y >= SCREEN_HEIGHT ||
		worm[headIndex].y < 0 
		) {
		return Loss; // collision of head with edge
	}
	
	for( c = 0; c < headIndex; c++ ) {
		if( worm[c].x == worm[headIndex].x &&
			worm[c].y == worm[headIndex].y
			) { 
			return Loss; // collision of head with another point of the worm
		}
	}
	
	if( worm[headIndex].x == food.x &&
		worm[headIndex].y == food.y
		) {
		return AteFood; // collision of head with food
	}
	
	return Nothing;
}

void display_separator() {
	ubyte c;
	for( c = 0; c < SCREEN_WIDTH; c++ ) {
		DrawLCDPix( c, SCREEN_HEIGHT );
	}
}

// for some reason, division by anything but a power of 2 (compiled
// straight into a shift most likely) did not work for me, so I wrote
// my own...
ubyte div8( ubyte numerator, ubyte denominator ) {
	ubyte result = 0;
	while( numerator >= denominator ) {
		result++;
		numerator -= denominator;
	}
	return result;
}
ubyte mod8( ubyte numerator, ubyte denominator ) {
	while( numerator >= denominator ) {
		numerator -= denominator;
	}
	return numerator;
}

void display_score( ubyte headIndex ) {
	#define PADDING 0
	DrawString( PADDING, 27, "GOAL:" );
	DrawChar( PADDING+24, 27, '0'+div8(GOAL-headIndex+1, 10) );
	DrawChar( PADDING+28, 27, '0'+mod8(GOAL-headIndex+1, 10) );
}

void relocate_food( Point* food ) {
	// the "1+" and "-2" are so that food can't spawn right at the edge of the screen
	// the convenience is that it also eliminates unfair corner spawns
	food->x = 1 + mod8( (g_frame*3+5), SCREEN_WIDTH - 2 );
	food->y = 1 + mod8( (g_frame*2+1), SCREEN_HEIGHT - 2 );
}

void handle_input( byte isLengthOne, byte* dx, byte* dy, ubyte input ) {
	
	byte canMove; // used to determine if the worm can move in a certain direction
				  // this is because if it's longer than a pixel, it cannot change
				  // direction in the opposite direction
	
	// comparison to -1 doesn't work properly; this little hack works around it
	byte temp;

	canMove = isLengthOne || *dy != 1;
	if( (input & PAD_UP) && canMove ) {
		*dx = 0;
		*dy = -1;
		return;
	}

	temp = *dy;
	temp++;
	canMove = isLengthOne || temp != 0;
	if( (input & PAD_DOWN) && canMove ) {
		*dx = 0;
		*dy = 1;
		return;
	}
	
	canMove = isLengthOne || *dx != 1;
	if( (input & PAD_LEFT) && canMove ) {
		*dx = -1;
		*dy = 0;
		return;
	}
	
	temp = *dx;
	temp++;
	canMove = isLengthOne || temp != 0;	
	if( (input & PAD_RIGHT) && canMove ) {
		*dx = 1;
		*dy = 0;
		return;
	}
}

void extend_worm( Point* worm, ubyte* headIndex, byte dx, byte dy ) {
	*headIndex = *headIndex + 1;
	worm[*headIndex].x = worm[*headIndex-1].x + dx;
	worm[*headIndex].y = worm[*headIndex-1].y + dy;
}

void game() {
	Point worm[MAX_WORM_LENGTH];
	ubyte input = 0;
	ubyte oldInput = 0;
	ubyte c;
	byte dx = 1; // movement delta x
	byte dy = 0; // movement delta y
	byte oldDx, oldDy;
	ubyte headIndex; // head of the worm
	
	Point food;
	Outcome outcome;
	
	worm[0].x = SCREEN_WIDTH / 2;
	worm[0].y = SCREEN_HEIGHT / 2;
	worm[1].x = worm[0].x + dx;
	worm[1].y = worm[0].y;
	headIndex = 1;
	
	relocate_food( &food );
	
	while( 1 ) {		
		FastLCDClear();
		g_frame++;
		
		move_worm( &worm[0], headIndex, dx, dy );
		display_worm( &worm[0], headIndex );
		display_food( &food );
		display_separator();
		display_score( headIndex );
		
		outcome = check_outcome( &worm[0], headIndex, food );
		
		switch( outcome ) {
			case Loss:
				return;
			case AteFood:
				if( GOAL - headIndex == 0 ) {
					victory();
					return;
				}
				relocate_food( &food );
				extend_worm( &worm[0], &headIndex, dx, dy );
				break;
			case Nothing:
				break;
		}
		
		VSync(1);
		
		// add delay to slow down action, checking input every iteration
		// delay decreases with score
		for( c = 0; c < NUM_FRAMES_DELAY - div8( headIndex, 3 ); c++ ) {
			oldInput = input;
			input = PAD_STATUS;
			if( ( input & PAD_ANY ) && oldInput != input ) {
				oldDx = dx; 
				oldDy = dy;
				handle_input( headIndex == 0, &dx, &dy, input );
				if( dx != oldDx || dy != oldDy ) {
					break;
				}
			}
			delay();
		}
	}
}

void main() {
	PocketInit();

	while( 1 ) {
		menu();
		game();
	}

	PocketExit();
}
